/* 
 * File:   main.c
 * Author: Pavel
 *
 * Created on 21. prosince 2015, 14:08
 */

//#include <stdio.h>
#include <stdlib.h>
#include <xc.h>
#include <string.h>
#include "rkods.h"
#include "font.c"
#include <setjmp.h>



// CONFIG1 
#pragma config FOSC = INTOSC    // Oscillator Selection Bits (INTOSC oscillator: I/O function on CLKIN pin) 
#pragma config WDTE = OFF       // Watchdog Timer Enable (WDT disabled) 
#pragma config PWRTE = ON      // Power-up Timer Enable (PWRT disabled) 
#pragma config MCLRE = ON       // MCLR Pin Function Select (MCLR/VPP pin function is MCLR) 
#pragma config CP = OFF         // Flash Program Memory Code Protection (Program memory code protection is disabled) 
#pragma config BOREN = OFF      // Brown-out Reset Enable (Brown-out Reset disabled) 
#pragma config CLKOUTEN = OFF   // Clock Out Enable (CLKOUT function is disabled. I/O or oscillator function on the CLKOUT pin) 

// CONFIG2 
#pragma config WRT = OFF        // Flash Memory Self-Write Protection (Write protection off) 
#pragma config STVREN = ON      // Stack Overflow/Underflow Reset Enable (Stack Overflow or Underflow will cause a Reset) 
#pragma config BORV = LO        // Brown-out Reset Voltage Selection (Brown-out Reset Voltage (Vbor), low trip point selected.) 
#pragma config LPBOR = OFF      // Low-Power Brown Out Reset (Low-Power BOR is disabled) 
#pragma config LVP = ON        // Low-Voltage Programming Enable (High-voltage on MCLR/VPP must be used for programming) 
#pragma config PPS1WAY = 0FF 


#define BFS 128           // kruhov buffer 
#define RS PORTAbits.RA0  //display text - command  
#define EN PORTAbits.RA1  //display enable 

#define CMD 0  //command
#define DAT 1    //data



const char * const label[] = {"SPI", "I2C 0x", "Bd ","LCD 1602 - 2004 ", "Hz", "Kmitoet"};

const char * const BaudText[] = {
    "300", "1200", "2400", "4800", "9600", "19200", "38400", "57600", "115200"
};
    

static const char W1250[] = {
    0xE1, 0xE8, 0xEC, 0xED, 0xF8, 0x9A, 0xB9, 0xFA, 0xFD, 0xF2, 0xB0, 0xB5, //          m
    8,       9,  0xA,  0xB,  0xC,  0xD, 0xD,  0xE,  0xF, 0xEE, 0xDF, 0xE4
};

static const char Baud[] = {
    0x04, 0x0D, 0x40, 0x03, 0xA0, 0x01, 0xCF, 0x00, 0x67, 0x00, 0x33, 0x00, 0x19, 0x00, 0x10, 0x00, 0x08, 0x00
}; //300,       1200,       2400,       4800,       9600,       19200,      38400,      57600,      115200

//static const char ASCII[];
//static const char bargr[][24];
void FreqCnt(void);

char buffer[BFS];
char str1[10];
char str2[26];
char wptr = 0;
char rptr = 0;
char dsize = 0;
char comm = 0;
char pchr = 0;
char bt = 0;
char binp = 0;
char addr = 0;
char chr = 0,  count = 0, tf=0;
char tml=0, tmh=0, vol=0, voh=0, vddl=0, vddh=0; 
char rc2mod = 0;
char rdch[34];

char  tup, rng=0, freq=0, tflg=0;
unsigned long tmrup;


void BaudSet(char baud) {
    char bd = baud - 1;
    SPBRGL = Baud[bd << 1];
    SPBRGH = Baud[(bd << 1) + 1];

}

void Delay(char ms) { //ms
    char tms = ms;
    while (tms) {
        _delay(4000);
        tms--;
    }
}

void Sdelay(char dl) {
    char ddl = dl;
    while (ddl) {
        _delay(32); //8us
        ddl--;
    }
}

void Dport(char dch) {

    if (dch & 1)PORTAbits.RA5 = 1;
    else PORTAbits.RA5 = 0;
    if (dch & 2)PORTAbits.RA4 = 1;
    else PORTAbits.RA4 = 0;
    if (dch & 4)PORTCbits.RC5 = 1;
    else PORTCbits.RC5 = 0;
    if (dch & 8)PORTCbits.RC4 = 1;
    else PORTCbits.RC4 = 0;
}

void LCD_Start(char ch) {
    Dport(ch);
    EN = 0;
    EN = 1;
    if (ch == 3) Sdelay(200); //1,5 ms
    else Sdelay(4); //32us
}

void LCD_Chr(char dchr, char dc) {
    char i;
    
    if (dc && (dchr > 0x7F) ) {
        for (i = 0; i < 12; i++)if (dchr == W1250[i]) {
                dchr = W1250[i + 12];
                break;
            }
    };
 
    RS = dc; //RS = 1(data)
    Dport((dchr & 0xF0) >> 4);
    EN = 1;
    EN = 0;
    Dport(dchr & 0x0F);
    EN = 1;
    EN = 0;
   
    if (dc == CMD && (dchr < 8)) Sdelay(200); // __delay_ms(2);
    else Sdelay(4); //40us
    EN = 1;

}

void LCD_Init(void) {
    Delay(40);

    EN = 1;
    RS = 0;
    LCD_Start(3); //ekej > 4,1ms
    LCD_Start(3);
    LCD_Start(3);
    LCD_Start(2);

    LCD_Chr(0x28, CMD); //dva rdky
    LCD_Chr(0x10, CMD); //cursor left
    LCD_Chr(0x0E, CMD); //cursor on, disp on
    LCD_Chr(0x06, CMD); //incrementace
    LCD_Chr(0x01, CMD); //smazat LCD

}

void LCD_Strx(char *str) {
    while (*str != 0)
        LCD_Chr(*str++, DAT);
}

char MemRead() {

    PMADRL = 0xE0; //Store LSB of address
    PMADRH = 0xF; //Store MSB of address
    PMCON1bits.CFGS = 0; //Do not select Configuration Space
    PMCON1bits.RD = 1; //Initiate read
    asm("NOP"); //Ignored (Figure 10-2)
    asm("NOP"); //Ignored (Figure 10-2)
    //meml = PMDATL ; //Get LSB of word
    //memh = PMDATH ; // Get MSB of word
    //str1[7]= meml;// + 0x30;
    return PMDATL;
}

void Unlock() {
    PMCON2 = 0x55; // Write 55h
    PMCON2 = 0xAA; //Write AAh
    PMCON1bits.WR = 1; // Set WR bit to begin write
    asm("NOP");
    asm("NOP");
}

void MemWrite(char dat) {
    GIE = 0; //Disable ints so required sequences will execute properly
    PMADRH = 0xF; //Store MSB of address
    PMADRL = 0xE0; //Store LSB of address

    PMCON1bits.CFGS = 0; //Do not select Configuration Space
    PMCON1bits.FREE = 0; //not Specify an erase operation
    PMCON1bits.LWLO = 1; // Only Load Write Latches
    PMCON1bits.WREN = 1; //Initiate Enable writes

    PMDATL = dat;
    PMDATH = 0;
    Unlock();
    PMCON1bits.LWLO = 0; //No more loading latches - Actually start Flash program
    Unlock();
    PMCON1bits.WREN = 0; //Disable writes
}

void MemEras() {
    GIE = 0; //Disable ints so required sequences will execute properly
    PMADRL = 0xE0;
    PMADRH = 0xF;
    PMCON1bits.CFGS = 0; //Not configuration space
    PMCON1bits.FREE = 1; //Specify an erase operation
    PMCON1bits.WREN = 1; //Enable writes
    Unlock();
    PMCON1bits.WREN = 0; //Disable writes
}

char butn() {
    char but = 0;
    TRISAbits.TRISA0 = 1;
    TRISAbits.TRISA2 = 1;
    TRISAbits.TRISA5 = 1;
    TRISAbits.TRISA4 = 1;
    TRISCbits.TRISC5 = 1;
    TRISCbits.TRISC4 = 1;
    Sdelay(2); //16us
    if (!PORTAbits.RA0)but |= 0x20;
    if (!PORTAbits.RA2)but |= 0x10;
    if (!PORTAbits.RA5)but |= 8;
    if (!PORTAbits.RA4)but |= 4;
    if (!PORTCbits.RC5)but |= 2;
    if (!PORTCbits.RC4)but |= 1;

    TRISAbits.TRISA0 = 0;
    TRISAbits.TRISA2 = 0;
    TRISAbits.TRISA4 = 0;
    TRISAbits.TRISA5 = 0;
    TRISCbits.TRISC5 = 0;
    TRISCbits.TRISC4 = 0;

    return but;
}

void push_char(char c) {
    wptr++;
    wptr &= (BFS - 1);
    buffer[wptr] = c;
    dsize++;
}

/*
char buffer_full(void) {
    return rptr == wptr && dsize == BFS;
}

char buffer_empty(void) {
    return rptr == wptr && dsize == 0;
}
 */
char pull_char(void) {
    rptr++;
    rptr &= (BFS - 1);
    dsize--;
    return buffer[rptr];
}


void interrupt IntSR() {
  
   static char rpt=0, wpt=0; 

    if (PIR1bits.SSP1IF) {
        PIR1bits.SSP1IF = 0;

        if (addr > 0x0F) {   //I2C
           SSP1CON1bits.CKP = 0;
             if (SSP1STATbits.R_nW) {  
                 if(!ACKSTAT){
                 SSP1BUF = (char) rdch[rpt];                                                                 
                rpt++; if(rpt>31)rpt=0;}
                 else rpt=0;  
                 _delay(10); 
             }
             else {
                if (SSP1STATbits.D_nA == 0) { //addr                     
                   chr = SSP1BUF; 
                } else {   
                   
                   chr = SSP1BUF; rpt=0; 
                                       
                    switch(chr){
                        case RSET: asm("RESET");
                        break;
                        case BUFF:  rdch[0] = dsize;
                        break;
                        case BUTN:  rdch[0] = butn(); // + 0x30;
                        break;
                        case TEMP:  rdch[0] = tml; rdch[1]= tmh;// adc - temper.                      
                        break;
                        case VDD:  rdch[0] = vddl; rdch[1]= vddh;  // Vdd                        
                        break;
                        case VOLT:  rdch[0] = vol; rdch[1]= voh;  // adc - RC3.                        
                        break;
                        case RTXT:  if(freq) strcpy(rdch,str2);    else strcpy(rdch,label[3]);    //text 
                                    
                        break;            
                        default:    push_char(chr);
                    } //switch
                  
                }
                }
            SSP1CON1bits.CKP = 1;
             
        }//if addr
        else if (addr == 0) {  //SPI
          
                if(rpt==0){ chr = SSP1BUF;
                    switch(chr){
                        case RSET: asm("RESET");
                        break;
                        case BUFF:   rdch[0] = dsize; wpt=0; rpt = 2;
                        break;
                        case BUTN: rdch[0] = bt;   wpt=0; rpt = 2; 
                        break;
                        case TEMP: rdch[0] = tml; rdch[1] = tmh; wpt=0; rpt=3; // adc - temper.                                                
                        break;
                        case VDD: rdch[0] = vddl; rdch[1] = vddh; wpt=0; rpt = 3;  //vdd 
                        break;
                        case VOLT: rdch[0] = vol; rdch[1] = voh; wpt=0; rpt = 3;  //napeti na RC3
                        break;
                        case RTXT:  strcpy(rdch,label[3]); wpt=0;  rpt = 16;   //text                    
                        break;                        
                        default:    push_char(chr); 
                        break;
                            
                    } //switch
                }
             if(rpt>0) {rpt--; SSP1BUF = rdch[wpt++];  }
        
        }

    } //if SSPIF
    if (RCIF) {    //UART
        chr = RC1REG;
        
                    switch(chr){
                        case RSET: asm("RESET");
                        break;
                        case BUFF:  TX1REG = dsize; 
                        break;
                        case BUTN: TX1REG = bt; //butn() + 0x30; 
                        break;
                        case TEMP: TX1REG = tml; while(!TRMT); TX1REG = tmh; // adc - temper.                         
                        break;
                        case VDD: TX1REG = vddl; while(!TRMT); TX1REG = vddh;   // vdd                         
                        break;
                        case VOLT: TX1REG = vol; while(!TRMT); TX1REG = voh;   // adc - RC3.                         
                        break;
                        case RTXT:  if(freq) strcpy(rdch,str2);    else strcpy(rdch,label[3]); 
                                    rpt=0;   //text 
                        while(rpt<16){ while(!TRMT);  TX1REG = rdch[rpt++];}                        
                        break;
            
            
                        default: push_char(chr);  
                    } //switch   
    }
  // }//if freq==0

    if (TMR2IF) {
        TMR2IF = 0;
        if (freq) {
            tf++;
            if (tf == 100)
                tf = 0;
            if (tf == 1) {
                TMR1ON = 1; //otevre Tmr1
            } else if (tf == 41) {
                asm("NOP");
                asm("NOP");
                asm("NOP");
                TMR1ON = 0; // zavre Tmr1             
                tflg = 1;
            };
        };

        count++;
        if (count >= 20) {
            count = 0;
            binp = 1;
        }

    }//if tmr2if

    if (TMR1IF) {
        tup++;
        TMR1IF = 0;
    }

} //int

void OpenI2C() {
    SSP1STAT &= 0x3F; // power on state               
    SMP=1;  //standard speed mode
    CKE=1;  //enable SMBus specification
    SSP1CON1 = 0x16; // select I2C
    SEN = 1;  // clock stretching enabl.
    CKP = 1; // enable clock
    BOEN = 1; //ignoruje BF peteen  
    PIE1bits.SSP1IE = 1;
    SSP1CON1bits.SSPEN = 1;

}

void OpenSPI() {
    SSP1STAT &= 0x3F; // power on state
    SMP = 0; //spi slave
    CKE = 1; //clock polar   1->0    
    SSP1CON1 = 4; //SS enab, 5 = disab. 
    CKP = 0; // idle clk = 0
    BOEN = 1; //ignoruje BF peteen
    PIE1bits.SSP1IE = 1;
    SSP1CON1bits.SSPEN = 1;
}

void OpenUSART() {
    SPEN = 1;
    SYNC = 0;
    CREN = 1;
    TXEN = 1;
    BRGH = 0;
    BRG16 = 1;
    BaudSet(addr); //5 = 9600
    RCIE = 1;
}

void Fonts(char beg, char len, char *font) {
    char i;
    LCD_Chr(ADRF + beg, CMD); //adresa ram
    for (i = 0; i < len; i++)
        LCD_Chr(font[i], DAT);
}

int main() {
    OSCCON = 0x78; //16 MHz
    OPTION_REGbits.nWPUEN = 0;
    PORTA = 0x0;
    PORTC = 0;
    LATA = 0;
    LATC = 0;
    ANSELA = 0;
    ANSELC = 0;
    TRISA = 0x0;
    TRISC = 0x0B; // Port C C0, C1 C3 input

    WPUA0 = 1;
    WPUA2 = 1;
    WPUA4 = 1;
    WPUA5 = 1;
    WPUC0 = 1;
    WPUC1 = 1;
    WPUC3 = 1;
    WPUC4 = 1;
    WPUC5 = 1;

    Delay(100);
    bt = butn();

    if (nRMCLR == 0 && nRI != 0 && (bt & 0x20)) {
        if (!(bt & 0x10) && (bt & 0xF) == 0) { //SPI
            addr = 0;
            chr = 1;
        }//SPI

        else if (!(bt & 0x10) && (bt & 0xF) > 0) { //I2C                 
            addr = bt << 4;
            chr = 1;
        }; //I2C

        if ((bt & 0x10) && (bt & 0xF) > 0) { //UART 
            addr = bt & 0xF;
            if (addr > 9)addr = 9;
            chr = 1;
        }

        if ((bt & 0x10) && (bt & 0xF) == 0) { //FreqCnt
            freq = 1;
        }

        if (chr) {
            MemEras();
            MemWrite(addr);
        }
    }
    
  
    nRI = 1; //reset instruction set bit
    TMR2IE = 1;
    PWM3CON = 0xF;
    PR2 = 249;
    PWM3DCL = 0;
    PWM3DCH = 0x80;
    T2CON = 0x4E; //x6;

    PWM3EN = 1;
    //  PWM3OE = 1;  // pro PIC16F1503
    //  PWM3CONbits.

    EN = 1;
    LCD_Init();

    Fonts(0,64,(char *)ASCII);
    
 asm("beg"); 
 
    addr = MemRead();
   /*---------------- Peripheral Pin Select (PPS) -------------------*/
    GIE = 0;
    PPSLOCK = 0x55;
    PPSLOCK = 0xAA;
    PPSLOCKED = 0; //enable PPS
    RA2PPS = 0x0E; // RA2 jako PWM3OUT 
    if (addr == 0)RC2PPS = 0x12; //RC2 jako SDO
    else {RC2PPS = 0;   } //RC2 jako I/O
    if (addr < 0x0F)RC0PPS = 0x14; //RC0 jako TX 
    else RC0PPS = 0x10; //RC0 jako SCLK 
    RC1PPS = 0x11; //RC1 jako SDA    
    RXPPS = 0x11; // RC1 jako RX
    
   if(freq) T1CKIPPS = 0x13;  //RC3 jako T1CLK
      
    PPSLOCKED = 1; //dis. PPS
   /*-------------------------------------------------*/ 
    
    if(freq) {T1CON= 0x84; T1GCON = 0; }   //init TMR1 

    if (addr == 0) {
        strcpy(str1, label[0]);
        OpenSPI();
    } else if (addr > 0x0F) {
        SSP1ADD = addr;
        strcpy(str1, label[1]);
        str1[6] = ((addr >> 5) + 0x30);
        str1[7] = ((addr >> 1) & 0x8) + 0x30;
        str1[8] = 0;
        OpenI2C();
    } else {
        strcpy(str1, label[2]);
        strcat(str1, BaudText[addr - 1]);
        OpenUSART();
    }
    if(freq!=5){
    LCD_Chr(0x1, CMD);
    Delay(255);
    LCD_Strx(str1);

    Delay(255);
    Delay(255);

    LCD_Chr(1, CMD);
   // if(freq){Delay(255); LCD_Strx((char *)label[5]); };
    }
    ADCON1 = 0xD0; //fosc/16
    FVREN = 1;
    FVRCONbits.ADFVR0 = 1; 
    TSEN = 1;
    TSRNG = 0;
    TMR2IE = 1;
    if(freq) TMR1IE = 1; else {TMR1IE=0; T1CON = 0;}

    PEIE = 1;
    GIE = 1;

    comm = 0; 
  
    while (1) {
      
        if (!(rptr == wptr && dsize == 0)) {
            
            pchr = pull_char();
            if (comm) {
                if (comm == BRIG) {
                    if(freq)OSCTUNE = pchr;   else PWM3DCH = pchr;}
                else if(comm == DMOD){
                    switch(pchr){
                        case RC2L: RC2 = 0; rc2mod = 0;
                        break;
                        case RC2H: RC2 = 1; rc2mod = 0;
                        break;
                        case RC2B: RC2 = 0; rc2mod = 1;
                        break; 
                        
                        case FNT0: Fonts(0,64,(char *)ASCII);                           
                        break;                        
                        case FNT1: //Fonts(40,24,(char *)bargr[0]);                                                                       
                        case FNT2: //Fonts(40,24,(char *)bargr[1]);                                                                           
                        case FNT3: Fonts(40,24,(char *)bargr[(pchr-1)&3]);                           
                        break;
                        case FQE1:
                        case FQE2:
                        case FQE3:
                        case FQE4: /*LCD_Chr(1,CMD);*/
                        case FQEN: freq=pchr&7;  asm("goto beg");
                        break;
                        case FOFF: freq=0; /*LCD_Chr(1,CMD);*/ asm("goto beg");
                        break;
                        
                    }
                }
                
                    else if (comm == ERAZ){
                       
                        char i=0;
                       if(pchr <0x94)i=0x93-pchr;
                       else if (pchr < 0xA8)i= 0xA7-pchr;
                       else if( pchr < 0xD4)i = 0xD3-pchr;
                       else i = 0xE7 - pchr;
                           
                       LCD_Chr(pchr, CMD); while(i--)LCD_Chr(0x20,DAT); LCD_Chr(pchr, CMD); 
                    }  
                    else if(comm==BARG){
                        signed char bx;
                        char i=0;
                        for(bx=pchr; bx>2; bx-=3){LCD_Chr(0xF,DAT); i++;}
                        if(bx==2){LCD_Chr(0xE,DAT); i++;}
                        else if(bx==1){LCD_Chr(0xD,DAT); i++;}
                        for(; i<20; i++)LCD_Chr(0x20,DAT);
                        
                    }
            else if(comm==CMND)LCD_Chr(pchr, CMD);
                comm = 0;
        
               
            }
                else {
                     
                if (pchr == CMND || pchr == BRIG || pchr == ERAZ || pchr == DMOD || pchr==BARG)comm = pchr;
                else if (comm == 0) LCD_Chr(pchr, DAT);
            
                }
      } // if dsize

        
        if (binp) {
            binp = 0; 
            
          bt = butn();
          if(rc2mod){if(bt)RC2=0; else RC2=1;};  //je-li bt, RC2 = 0
          ADCON0 = 0x75; // adc - temper.
                        Sdelay(25); GO = 1; while (GO); 
                        tml = ADRESL; tmh = ADRESH;
          if(freq==0){            
          if(addr){             
          ANSC3 = 1; ADCON0 = 0x1D; // adc - RC3.
                        Sdelay(25); GO = 1; while (GO);
                        vol = ADRESL; voh = ADRESH; ANSC3 = 0; 
          }     
          }
             ADCON0 = 0x7D; //           VDD a VREF
                        Sdelay(25); GO = 1; while (GO);
                        vddl = ADRESL; vddh = ADRESH; //ANSC3 = 0;              
      //   }
        }
      
        if(freq && tflg){tflg=0; FreqCnt();}

    }//while

    return (0);
}

void FreqCnt() {

    unsigned long tml, tmh;
    char len = 0, row;
    char frac = '0';

    tml = TMR1L;
    tmh = TMR1H;
    tmrup = tup;

    tml += ((tmh << 8) + (tmrup << 16));
    TMR1L = TMR1H = tup = 0;
    TMR1IF = 0;

    ultoa(str2, tml >> 2, 10);

    while (str2[len] != 0)len++;

    if ((char) tml & 2)frac = '5';

    if (len > 2) {
        str2[len] = str2[len - 1];
        str2[len - 1] = str2[len - 2];
        str2[len - 2] = str2[len - 3];

        if (len > 5) {
            str2[len - 3] = str2[len - 4];
            str2[len - 4] = str2[len - 5];
            str2[len - 5] = '.';
            len -= 3;
            str2[len + 3] = 'M';
        }//if len>5
        else {
            str2[len - 2 ] = '.';
            str2[len + 3] = 'k';
            str2[len + 1] = frac;
        }
        str2[len + 2] = 0x20;
        strcpy(&str2[len + 4], label[4]);

    }//if len>2
    else {
        if (tml > 0) {
            str2[len] = frac;
        }//if tml
        else len--;
        str2[len + 1] = 0x20;
        strcpy(&str2[len + 2], label[4]); //Hz
    }

    len = 0;
    while (str2[len] != 0)len++;
    while ((len) < 15) {
        str2[len] = 0x20;
        len++;
    }

    str2[15] = 0;
    str2[16] = 0;
    if(freq<5){
        switch(freq){            
            case 2: row = ROW2;
            break;
            case 3: row = ROW3;
            break;
            case 4: row = ROW4;
            break;
            default: row = ROW1;
        }
        }
    LCD_Chr(row, CMD);
    LCD_Strx(str2);
    }


